@@ -65,6 +65,7 @@ extern "C" {
 #include <string.h>
 #include <stddef.h>
 #include <signal.h>
+#include <limits.h>
 #ifdef _OPENMP
     #include <omp.h>
 #else
@@ -78,30 +79,55 @@ extern "C" {
     #include <R.h>
     #include <R_ext/Rdynload.h>
     #include <R_ext/Print.h>
-    #define fprintf(f, message) REprintf(message)
+    #define fprintf(f, ...) REprintf(__VA_ARGS__)
     #define sparse_ix int
+    #undef USE_FLOAT
+#endif
+
+#ifndef USE_FLOAT
+    #ifndef real_t
+        #define real_t double
+    #endif
+    #define cblas_tdot cblas_ddot
+    #define cblas_taxpy cblas_daxpy
+    #define cblas_tscal cblas_dscal
+    #define cblas_tnrm2 cblas_dnrm2
+    #define cblas_tgemv cblas_dgemv
+#else
+    #ifndef real_t
+        #define real_t float
+    #endif
+    #define cblas_tdot cblas_sdot
+    #define cblas_taxpy cblas_saxpy
+    #define cblas_tscal cblas_sscal
+    #define cblas_tnrm2 cblas_snrm2
+    #define cblas_tgemv cblas_sgemv
 #endif
 
 /* https://www.github.com/david-cortes/nonneg_cg */
-typedef void fun_eval(double x[], int n, double *f, void *data);
-typedef void grad_eval(double x[], int n, double grad[], void *data);
-typedef void callback(double x[], int n, double f, size_t iter, void *data);
-int minimize_nonneg_cg(double x[], int n, double *fun_val,
+typedef void fun_eval(real_t x[], int n, real_t *f, void *data);
+typedef void grad_eval(real_t x[], int n, real_t grad[], void *data);
+typedef void callback(real_t x[], int n, real_t f, size_t iter, void *data);
+int minimize_nonneg_cg(real_t x[], int n, real_t *fun_val,
                        fun_eval *obj_fun, grad_eval *grad_fun, callback *cb, void *data,
-                       double tol, size_t maxnfeval, size_t maxiter, size_t *niter, size_t *nfeval,
-                       double decr_lnsrch, double lnsrch_const, size_t max_ls,
-                       double *buffer_arr, int nthreads, int verbose);
+                       real_t tol, size_t maxnfeval, size_t maxiter, size_t *niter, size_t *nfeval,
+                       real_t decr_lnsrch, real_t lnsrch_const, size_t max_ls,
+                       bool limit_step, real_t *buffer_arr, int nthreads, int verbose);
 /* Data struct to pass to nonneg_cg */
 typedef struct fdata {
-    double *F;
-    double *Fsum;
-    double *X;
+    real_t *B;
+    real_t *Bsum;
+    real_t *Xr;
     sparse_ix *X_ind;
     sparse_ix nnz_this;
-    double l2_reg;
-    double w_mult;
+    real_t l2_reg;
+    real_t w_mult;
+    int k;
 } fdata;
 
+/* TNC */
+#include "tnc.h"
+
 /* BLAS functions */
 
 #ifdef _FOR_PYTHON
@@ -112,11 +138,12 @@ typedef struct fdata {
     typedef enum CBLAS_TRANSPOSE {CblasNoTrans=111, CblasTrans=112, CblasConjTrans=113, CblasConjNoTrans=114} CBLAS_TRANSPOSE;
     typedef enum CBLAS_UPLO      {CblasUpper=121, CblasLower=122} CBLAS_UPLO;
     typedef CBLAS_ORDER CBLAS_LAYOUT;
-    double cblas_ddot(const int n, const double *x, const int incx, const double *y, const int incy);
-    void cblas_daxpy(const int n, const double alpha, const double *x, const int incx, double *y, const int incy);
-    void cblas_dscal(const int N, const double alpha, double *X, const int incX);
-    void cblas_dgemv(const CBLAS_ORDER order,  const CBLAS_TRANSPOSE TransA,  const int m, const int n,
-         const double alpha, const double  *a, const int lda,  const double  *x, const int incx,  const double beta,  double  *y, const int incy);
+    real_t cblas_tdot(const int n, const real_t *x, const int incx, const real_t *y, const int incy);
+    void cblas_taxpy(const int n, const real_t alpha, const real_t *x, const int incx, real_t *y, const int incy);
+    void cblas_tscal(const int N, const real_t alpha, real_t *X, const int incX);
+    real_t cblas_tnrm2(const int n, const real_t *x, const int incx);
+    void cblas_tgemv(const CBLAS_ORDER order,  const CBLAS_TRANSPOSE TransA,  const int m, const int n,
+         const real_t alpha, const real_t  *a, const int lda,  const real_t  *x, const int incx,  const real_t beta,  real_t  *y, const int incy);
     #endif
 #endif
 
@@ -136,77 +163,98 @@ typedef struct fdata {
 /* Function prototypes */
 
 /* poismf.c */
-void sum_by_cols(double *restrict out, double *restrict M, size_t nrow, size_t ncol);
+void dscal_large(size_t n, real_t alpha, real_t *restrict x);
+void sum_by_cols(real_t *restrict out, real_t *restrict M, size_t nrow, size_t ncol);
 void adjustment_Bsum
 (
-    double *restrict B,
-    double *restrict Bsum,
-    double *restrict Bsum_user,
+    real_t *restrict B,
+    real_t *restrict Bsum,
+    real_t *restrict Bsum_user,
     sparse_ix Xr_indices[],
     sparse_ix Xr_indptr[],
     size_t dimA, size_t k,
-    double w_mult, int nthreads
+    real_t w_mult, int nthreads
 );
-void calc_grad_pgd(double *out, double *curr, double *F, double *X, sparse_ix *Xind, sparse_ix nnz_this, int k);
+void calc_grad_pgd(real_t *out, real_t *curr, real_t *F, real_t *X, sparse_ix *Xind, sparse_ix nnz_this, int k);
 void pg_iteration
 (
-    double *A, double *B,
-    double *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
+    real_t *A, real_t *B,
+    real_t *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
     size_t dimA, size_t k,
-    double cnst_div, double *cnst_sum,
-    double step_size, double w_mult, size_t npass,
-    double *buffer_arr, int nthreads
+    real_t cnst_div, real_t *cnst_sum, real_t *Bsum_user,
+    real_t step_size, real_t w_mult, size_t maxupd,
+    real_t *buffer_arr, int nthreads
+);
+void calc_fun_single(real_t a_row[], int k_int, real_t *f, void *data);
+void calc_grad_single(real_t a_row[], int k_int, real_t grad[], void *data);
+void calc_grad_single_w(real_t a_row[], int k_int, real_t grad[], void *data);
+int calc_fun_and_grad
+(
+    real_t *restrict a_row,
+    real_t *restrict f,
+    real_t *restrict grad,
+    void *data
 );
-void calc_fun_single(double x[], int n, double *f, void *data);
-void calc_grad_single(double x[], int n, double grad[], void *data);
 void cg_iteration
 (
-    double *A, double *B,
-    double *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
+    real_t *A, real_t *B,
+    real_t *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
+    size_t dimA, size_t k, bool limit_step,
+    real_t *Bsum, real_t l2_reg, real_t w_mult, size_t maxupd,
+    real_t *buffer_arr, real_t *Bsum_w, int nthreads
+);
+void tncg_iteration
+(
+    real_t *A, real_t *B,
+    real_t *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
     size_t dimA, size_t k,
-    double *Bsum, double l2_reg, double w_mult, size_t npass,
-    double *buffer_arr, double *Bsum_w, int nthreads
+    real_t *Bsum, real_t l2_reg, real_t w_mult, int maxupd,
+    real_t *buffer_arr, int *buffer_int,
+    real_t *zeros_tncg, real_t *inf_tncg,
+    real_t *Bsum_w, int nthreads
 );
 void set_interrup_global_variable(int s);
 
 /* main function */
+typedef enum Method {tncg = 1, cg = 2, pg = 3} Method;
 int run_poismf(
-    double *restrict A, double *restrict Xr, sparse_ix *restrict Xr_indptr, sparse_ix *restrict Xr_indices,
-    double *restrict B, double *restrict Xc, sparse_ix *restrict Xc_indptr, sparse_ix *restrict Xc_indices,
+    real_t *restrict A, real_t *restrict Xr, sparse_ix *restrict Xr_indptr, sparse_ix *restrict Xr_indices,
+    real_t *restrict B, real_t *restrict Xc, sparse_ix *restrict Xc_indptr, sparse_ix *restrict Xc_indices,
     const size_t dimA, const size_t dimB, const size_t k,
-    const double l2_reg, const double l1_reg, const double w_mult, double step_size,
-    const bool use_cg, const size_t numiter, const size_t npass, const int nthreads);
+    const real_t l2_reg, const real_t l1_reg, const real_t w_mult, real_t step_size,
+    const Method method, const bool limit_step, const size_t numiter, const size_t maxupd,
+    const int nthreads);
 
 /* topN.c */
 bool check_is_sorted(sparse_ix arr[], size_t n);
-void qs_argpartition(sparse_ix arr[], double values[], size_t n, size_t k);
+void qs_argpartition(sparse_ix arr[], real_t values[], size_t n, size_t k);
 int cmp_size_t(const void *a, const void *b);
 int cmp_argsort(const void *a, const void *b);
 int topN
 (
-    double *restrict a_vec, double *restrict B, int k,
+    real_t *restrict a_vec, real_t *restrict B, int k,
     sparse_ix *restrict include_ix, size_t n_include,
     sparse_ix *restrict exclude_ix, size_t n_exclude,
-    sparse_ix *restrict outp_ix, double *restrict outp_score,
+    sparse_ix *restrict outp_ix, real_t *restrict outp_score,
     size_t n_top, size_t n, int nthreads
 );
 
 /* llk_and_pred.c */
 void predict_multiple
 (
-    double *restrict out,
-    double *restrict A, double *restrict B,
+    real_t *restrict out,
+    real_t *restrict A, real_t *restrict B,
     sparse_ix *ixA, sparse_ix *ixB,
     size_t n, int k,
     int nthreads
 );
 long double eval_llk
 (
-    double *restrict A,
-    double *restrict B,
+    real_t *restrict A,
+    real_t *restrict B,
     sparse_ix ixA[],
     sparse_ix ixB[],
-    double *restrict X,
+    real_t *restrict X,
     size_t nnz, int k,
     bool full_llk, bool include_missing,
     size_t dimA, size_t dimB,
@@ -214,22 +262,22 @@ long double eval_llk
 );
 int factors_multiple
 (
-    double *A, double *B, double *A_old, double *Bsum,
-    double *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
+    real_t *A, real_t *B, real_t *A_old, real_t *Bsum,
+    real_t *Xr, sparse_ix *Xr_indptr, sparse_ix *Xr_indices,
     int k, size_t dimA,
-    double l2_reg, double w_mult,
-    double step_size, size_t niter, size_t npass,
-    bool use_cg,
+    real_t l2_reg, real_t w_mult,
+    real_t step_size, size_t niter, size_t maxupd,
+    Method method, bool limit_step,
     int nthreads
 );
 int factors_single
 (
-    double *restrict out, size_t k,
-    double *restrict A_old, size_t dimA,
-    double *restrict X, sparse_ix X_ind[], size_t nnz,
-    double *restrict B, double *restrict Bsum,
-    size_t npass, double l2_reg, double l1_new, double l1_old,
-    double w_mult
+    real_t *restrict out, size_t k,
+    real_t *restrict A_old, size_t dimA,
+    real_t *restrict X, sparse_ix X_ind[], size_t nnz,
+    real_t *restrict B, real_t *restrict Bsum,
+    int maxupd, real_t l2_reg, real_t l1_new, real_t l1_old,
+    real_t w_mult
 );
 
 #ifdef __cplusplus
